'''
Created on Apr 7, 2014

@author: vladimim
'''
#Modified by Adriana Reyes to read measurements from spreadsheet
#import comtypes.client as comClient
#import comtypes.gen.UniSimDesign as UD
import sys
import win32com.client as w32c
from win32com.client import gencache
gencache.EnsureModule('{707BF17F-353C-40B2-A5D2-26B20CE7DCFF}', 0, 1, 1)
# import UnisimLib as UD
# import numpy as np
import logging


class Hysys2py:
    def __init__(self, cell):
        self.currentValue = cell

class Unisim2py:
    """A class that load/connects to an open Hysys document through COM and attempts to convert it to NLP format
                            min f(x) s.t. lb<=x<=ub h(x)=0 g(x)<=0 """
    def __init__(self):
        try:
            # Start a logger
            logging.basicConfig(level=logging.DEBUG)
            self.logger = logging.getLogger(type(self).__name__)
            self.lastFeasibleObjectiveTerms = []
            self.lastFeasibleInputs = []
            self.isObjectiveMultiTerm = False
            # Attempt to load/connect to Hysys/Unisim
            # gencache.EnsureModule('{707BF17F-353C-40B2-A5D2-26B20CE7DCFF}', 0, 1, 1)
            self.hyApp = w32c.Dispatch("UnisimDesign.Application")
            self.logger.debug("...Object initialized "+ repr(self.hyApp))
            self.hyCase = self.hyApp.ActiveDocument
            self.hyCase.Solver.CanSolve = True
            self.logger.debug("...Simulation loaded and running" + repr(self.hyCase))
            self.hyFlowsheet = self.hyCase.Flowsheet
            self.logger.debug("...Flowsheet loaded " + repr(self.hyFlowsheet))
            self._loadInputsAnsUnits()
            self.logger.debug("...Inputs references and their corresponding measurement units loaded. ")
            self._loadInitialValues()
            self.logger.debug("...References for the initial input values loaded. ")
            self._loadInputsBounds()
            self.logger.debug("...Inputs bounds references loaded. ")
            self._loadInequalityConstraints()
            self.logger.debug("...Inequality constraints references loaded")
            self._loadEqualityConstraints()
            self.logger.debug("...Equality constraints references loaded")
            self._loadObjectiveFunctionTerms()
            self.logger.debug("...Objective function reference loaded")
            self._loadParametersAndUnits()
            self.logger.debug("...Parameter references loaded")
            self._loadMeasurementsAndUnits()
            self.logger.debug("...Measurements references loaded")
            print "MAKE SURE THAT THE FLOWSHEET HAS CONVERGED BEFORE RUNNING THE PROGRAM!"
            print "PRINT THE NLP FORMULATION AND CHECK THAT EVERYTHING IS CORRECT"
        except Exception:
            self.logger.exception("Failed to initialize the Hysys2Py ", exc_info=True)
            sys.exit(1)

    def getLoadedSimulationName(self):
        return self.hyCase.FullName

    def _testFunction(self):
        print repr(self.hyFlowsheet.Operations)
        print repr(self.hyFlowsheet.Operations.Item("Inputs"))
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2")
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").AttachmentType
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").Units
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").VariableName
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").VariableType
        print repr(self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable)
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellValue
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellText
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").AttachedObjectName
        test1 = self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellValue
        print test1
        self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable.Value = 258.5
        test1 = 295
        print test1, self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellValue

        inputDict = {'input1': self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable}

        print inputDict["input1"]
        inputDict["input1"].Value = 260
        print inputDict["input1"]
        self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable.Value = 259
        print inputDict["input1"]
        inputDict["input1"].Value = 261
        print inputDict["input1"]
        test2 = self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable
        print test2.Value
        test2.Value = 262
        print self.hyFlowsheet.Operations.Item("Inputs").Cell("A2").CellVariable, inputDict["input1"]

    def _loadInputsAnsUnits(self):
        self.inputsList = list()
        self.unitsList = list()
        index = 1
        # Store references to Input cells
        while self.hyFlowsheet.Operations.Item("Inputs").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
            # Create lists
            self.inputsList.append(self.hyFlowsheet.Operations.Item("Inputs").Cell("A{0}".format(index+1)).CellVariable)
            self.unitsList.append(str(self.hyFlowsheet.Operations.Item("Inputs").Cell("E{0}".format(index+1)).CellText))
            index += 1
        # Save the number of inputs
        self.numberOfInputs = index-1
        if self.numberOfInputs<1:
            self.logger.debug("Can't find any inputs !")
            raise

    def _loadParametersAndUnits(self):
        self.parameterList = list()
        self.parameterUnitsList = list()
        index = 1
        # Store references to Input cells
        while self.hyFlowsheet.Operations.Item("Parameters").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
            # Create lists
            self.parameterList.append(self.hyFlowsheet.Operations.Item("Parameters").Cell("A{0}".format(index+1)).CellVariable)
            self.parameterUnitsList.append(str(self.hyFlowsheet.Operations.Item("Parameters").Cell("E{0}".format(index+1)).CellText))
            index += 1
        # Save the number of inputs
        self.numberOfParameters = index-1

    def _loadInputsBounds(self):
        self.upperBoundsList = list()
        self.lowerBoundsList = list()
        index = 1
        # Store references to Input cells
        while self.hyFlowsheet.Operations.Item("Inputs").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
            # Create lists
            self.lowerBoundsList.append(self.hyFlowsheet.Operations.Item("Inputs").Cell("B{0}".format(index+1)).CellVariable)
            self.upperBoundsList.append(self.hyFlowsheet.Operations.Item("Inputs").Cell("C{0}".format(index+1)).CellVariable)
            index += 1

    def _loadInitialValues(self):
        self.initialValuesList  = list()
        index = 1
        # Store references to Input cells
        while self.hyFlowsheet.Operations.Item("Inputs").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
            self.initialValuesList.append(self.hyFlowsheet.Operations.Item("Inputs").Cell("D{0}".format(index+1)).CellVariable)
            index += 1

    def _loadObjectiveFunctionTerms(self):
        self.objectiveFunctionTerms = list()
        index = 1
        try:
            # Load Objective
            while self.hyFlowsheet.Operations.Item("Objective").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
                self.objectiveFunctionTerms.append(self.hyFlowsheet.Operations.Item("Objective").Cell("A{0}".format(index+1)).CellVariable)
                index += 1
            self.numberOfObjectiveTerms = index-1
            if self.numberOfObjectiveTerm == 0:
                print "NO KNOWN OBJECTIVE FUNCTION TERMS !"
                raise
        except Exception:
            self.logger.exception("Failed to load objective function terms", exc_info=True)

    def _loadInequalityConstraints(self):
        self.inequalityConstraintsList = list()
        index = 1
        # Inequality Constraints
        try:
            while self.hyFlowsheet.Operations.Item("Constraints").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
                self.inequalityConstraintsList.append(self.hyFlowsheet.Operations.Item("Constraints").Cell("A{0}".format(index+1)).CellVariable)
                index += 1
            self.numberOfInequalityConstraints = index - 1
        except Exception:
            self.logger.exception("Failed to load the inequality constraints", exc_info=True)


    def _loadEqualityConstraints(self):
        self.equalityConstraintsList = list()
        index = 1
        # Inequality Constraints
        try:
            while self.hyFlowsheet.Operations.Item("Constraints").Cell("B{0}".format(index+1)).CellVariable.IsKnown:
                self.equalityConstraintsList.append(self.hyFlowsheet.Operations.Item("Constraints").Cell("B{0}".format(index+1)).CellVariable)
                index += 1
            self.numberOfEqualityConstraints = index - 1
        except Exception:
            self.logger.exception("Failed to load the equality constraints", exc_info=True)
    
           
    def _loadMeasurementsAndUnits(self):
        self.measurementsList = list()
        self.measurementUnitsList = list()
        index = 1
        # Store references to Input cells
        while self.hyFlowsheet.Operations.Item("Measurements").Cell("A{0}".format(index+1)).CellVariable.IsKnown:
            # Create lists
            self.measurementsList.append(self.hyFlowsheet.Operations.Item("Measurements").Cell("A{0}".format(index+1)).CellVariable)
            self.measurementUnitsList.append(str(self.hyFlowsheet.Operations.Item("Measurements").Cell("B{0}".format(index+1)).CellText))
            index += 1
        # Save the number of inputs
        self.numberOfMeasurements = index-1

    def getInputsValues(self, scaled=True):
        x = list()
        index = 0
        if not scaled:
            for cellVariable in self.inputsList:
                x.append(cellVariable.GetValue(self.unitsList[index]))
                index += 1
        else:
            for cellVariable in self.inputsList:
                x.append(self.scaleInputs(cellVariable.GetValue(self.unitsList[index]),index))
                index += 1
        return x
        for cellVariable in self.inputsList:
            x.append(cellVariable.GetValue(self.unitsList[index]))
            index += 1
        return x

    def setInputListValues(self,x):
        index = 0
        self.hyCase.Solver.CanSolve=False
        for cellVariable in self.inputsList:
            if (x[index]>=0 and x[index]<=1):
                cellVariable.SetValue(self.descaleInputs(x[index], index),self.unitsList[index])
            else:
                print "...Invalid inputs:", x
                sys.exit(1)
            index += 1
        self.hyCase.Solver.CanSolve=True
        print "x = ", x

    def setParametersValues(self,x):
        index = 0
        self.hyCase.Solver.CanSolve=False
        for cellVariable in self.parameterList:
            cellVariable.SetValue(x[index],self.parameterUnitsList[index])
            index += 1
        self.hyCase.Solver.CanSolve=True

    def getMeasurementsValues(self):
        x = list()
        index = 0
        for cellVariable in self.measurementsList:
            x.append(cellVariable.GetValue(self.measurementUnitsList[index]))
            index += 1
        return x
    
    def getParametersValues(self):
        x = list()
        index = 0
        for cellVariable in self.parameterList:
            x.append(cellVariable.GetValue(self.parameterUnitsList[index]))
            index += 1
        return x

    def getLowerBounds(self, scaled=True):
        x = list()
        index = 0
        if not scaled:
            for cellVariable in self.lowerBoundsList:
                x.append(cellVariable.GetValue())
                index += 1
        else:
            for cellVariable in self.lowerBoundsList:
                x.append(0)
                index += 1
        return x

    def getUpperBounds(self, scaled=True):
        x = list()
        index = 0
        if not scaled:
            for cellVariable in self.upperBoundsList:
                x.append(cellVariable.GetValue())
                index += 1
        else:
            for cellVariable in self.upperBoundsList:
                x.append(1)
                index += 1
        return x

    def getInitialInputValues(self, scaled=True):
        x = list()
        index = 0
        if not scaled:
            for cellVariable in self.initialValuesList:
                x.append(cellVariable.GetValue(self.unitsList[index]))
                index += 1
        else:
            for cellVariable in self.initialValuesList:
                x.append(self.scaleInputs(cellVariable.GetValue(self.unitsList[index]),index))
                index += 1
        return x

    def getInequalityConstraintValues(self):
        x = list()
        index = 0
        for cellVariable in self.inequalityConstraintsList:
            x.append(cellVariable.GetValue())
            index += 1
        return x

    def getEqualityConstraintValues(self):
        x = list()
        index = 0
        for cellVariable in self.equalityConstraintsList:
            x.append(cellVariable.GetValue())
            index += 1
        return x

    def getObjectiveFunctionTerms(self, inputs, isMultiTerm=False):
        self.setInputListValues(inputs)
        Jterms=list()
        if self.isObjectiveMultiTerm:
            feasibleSolution = True
            for index in range(self.numberOfObjectiveTerms):
                if self.objectiveFunctionTerms[index].IsKnown:
                    Jterms.append(self.objectiveFunctionTerms[index].Value)
                    print "Known term", self.objectiveFunctionTerms[index].Value
                else:
                    Jf = self.lastFeasibleObjectiveTerms[index]
                    xf = self.lastFeasibleInputs
                    Jc = Jf**(1+sum(abs(abs(xf)-abs(inputs))))
                    Jterms.append(Jc)
                    print "J{0} term of the objective function is unknown (probably simulation didn't converge)!".format(index)
                    print "Adding a penalty J{0}**(1+sum||xc|-|xf||)"
                    print "last known value of J{0} = {1}, current value of J{0} = {2} )".format(index,Jf,Jc)
                    feasibleSolution = False
            if feasibleSolution:
                self.lastFeasibleObjectiveTerms=Jterms
                self.lastFeasibleInputs = inputs
        else:
            Jterms.append(self.objectiveFunctionTerms[0].Value)
        print "J = sum({0}) = {1}".format(Jterms,sum(Jterms))
        return Jterms

    def scaleInputs(self, inputValue, inputIndex):
        return (inputValue - self.lowerBoundsList[inputIndex].Value)/(self.upperBoundsList[inputIndex].Value-self.lowerBoundsList[inputIndex].Value)

    def descaleInputsAll(self, inputValues):
        u_descaled = list()
        for inputIndex in xrange(0,len(inputValues)):
            u_descaled.append(self.lowerBoundsList[inputIndex].Value + inputValues[inputIndex] * (self.upperBoundsList[inputIndex].Value-self.lowerBoundsList[inputIndex].Value))
        return u_descaled

    def descaleInputs(self, inputValue, inputIndex):
        return self.lowerBoundsList[inputIndex].Value + inputValue * (self.upperBoundsList[inputIndex].Value-self.lowerBoundsList[inputIndex].Value)

    def printNLP(self):

        x = self.getInputsValues(scaled=False)
        x_scaled = self.getInputsValues()
        x0 = self.getInitialInputValues(scaled=False)
        x0_scaled = self.getInitialInputValues(scaled=True)
        lb = self.getLowerBounds(scaled=False)
        lb_scaled = self.getLowerBounds()
        ub = self.getUpperBounds(scaled=False)
        ub_scaled = self.getUpperBounds()
        param = self.getParametersValues()
        Jterms = self.getObjectiveFunctionTerms(x0_scaled)
        J = sum(Jterms)
        hx = self.getEqualityConstraintValues()
        gx = self.getInequalityConstraintValues()

        print "\npNLP format:\n"
        print "min f(x,p) starting from x0,p"
        print "x belong to R^{0}".format(self.numberOfInputs)
        print "subject to: lb <= x <= ub"
        print "            h(x,p) = 0"
        print "            g(x,p) <= 0\n"
        print "current values:"
        print "f(x)         = sum({0}) = {1}".format(Jterms,J)
        print "x0 (scaled)  = {0}".format(x0_scaled)
        print "x0           = {0}".format(x0)
        print "x (scaled)   = {0}".format(x_scaled)
        print "x            = {0}\n".format(x)
        print "parameter values"
        print "p = {0}\n".format(param)
        print "lb <= x <= ub   (scaled) {0} <= x_scaled <= {1}\n".format(lb_scaled[0],ub_scaled[0])

        for index in xrange(0,self.numberOfInputs):
            print "{index}: {lb} <= {x}[{unit}] <= {ub}".format(index=index, lb=lb[index], x=x[index], unit=self.unitsList[index], ub=ub[index])

        print "h(x) = {0}".format(hx)
        print "g(x) = {0}".format(gx)

def main():
    unisimDesignSimulation = Unisim2py()
    print "Loaded ! " +  unisimDesignSimulation.getLoadedSimulationName()
    unisimDesignSimulation.printNLP()

if __name__ == '__main__':
    main()


